import os.path

from jupyterhub.spawner import _quote_safe


class VariableMixin():

    def _pre_start(self):
        if self.user_options:
            trait_names = set(self.trait_names()) - {'user_options'}
            for k in trait_names.intersection(self.user_options.keys()):
                setattr(self, k, self.user_options[k])

    def _mixin_get_args(self, presentation_type):
        """Return the arguments to be passed after self.cmd

        Doesn't expect shell expansion to happen.

        Also adds self.args at the end in case specified by the config.
        """

        presentation_path = self.user_options.get('presentation_path', '')

        args = ['--destport=0']

        # jhsingle-native-proxy --destport $destport --authtype oauth voila `pwd` {--}port={port} {--}no-browser {--}Voila.base_url={base_url}/ {--}Voila.server_url=/ --port $port

        if self.ip:
            args.append('--ip=%s' % _quote_safe(self.ip))

        if self.port:
            args.append('--port=%i' % self.port)


        notebook_dir = '.'
        if self.notebook_dir:
            notebook_dir = self.format_string(self.notebook_dir)

        if presentation_path != '' and not '..' in presentation_path:
            # Should have been validated when dashboard created, but .. is particularly dangerous
            if presentation_path.startswith("/"):
                presentation_path = presentation_path[1:]
            notebook_dir = os.path.join(notebook_dir, presentation_path)


        voila_template = getattr(self, 'voila_template', '')

        if presentation_type == 'voila':

            args.extend(['python3', '{-}m','voila'])

            args.append(_quote_safe(notebook_dir))

            args.extend([
                '{--}port={port}',
                '{--}no-browser',
                '{--}Voila.base_url={base_url}/',
                '{--}Voila.server_url=/'
            ])

            if voila_template != '':
                args.append('='.join(('{--}template', voila_template)))

            if self.debug:
                args.append('{--}debug')

        elif presentation_type == 'streamlit':

            args.append('streamlit')

            if self.debug:
                args.append('{--}log_level=debug')

            args.append('run')

            args.append(_quote_safe(notebook_dir))

            args.extend([
                '{--}server.port={port}',
                '{--}server.headless=True',
                '{--}server.enableCORS=False'
            ])


            
        elif presentation_type == 'plotlydash':

            args.extend(['python3', '{-}m','plotlydash_tornado_cmd.main'])

            args.append(_quote_safe(notebook_dir))

            args.extend([
                '{--}port={port}'
            ])

            if self.debug:
                args.append('{--}debug')

        if self.debug:
            # jhsingle-native-proxy debug
            args.append('--debug')

        proxy_request_timeout = getattr(self, 'proxy_request_timeout', 0)
        if proxy_request_timeout:
            args.append('--request-timeout={}'.format(proxy_request_timeout))

        args.extend(self.args)
        return args

    def _get_presentation_type(self):
        if self.user_options and 'presentation_type' in self.user_options:
            return self.user_options['presentation_type']
        return ''

    def _mixin_get_env(self, env, presentation_type):
        if presentation_type == 'plotlydash':
            env['DASH_REQUESTS_PATHNAME_PREFIX'] = "{}/".format(self.server.base_url)
        return env


from systemdspawner import SystemdSpawner
from traitlets import Unicode, Integer

from .variablemixin import VariableMixin


class VariableSystemdSpawner(SystemdSpawner, VariableMixin):

    voila_template = Unicode(
        'materialstream',
        help="""
        --template argument to pass to Voila. Default is materialstream
        """,
    ).tag(config=True)

    proxy_request_timeout = Integer(
        0,
        help="""
        Request timeout in seconds that jhsingle-native-proxy should allow when proxying to the underlying process (e.g. Voila).
        The default of 0 means that no --request-timeout flag will be passed to jhsingle-native-proxy so it will use its own default.
        """,
    ).tag(config=True)

    async def start(self):
        self._pre_start()
        return await super().start()

    def get_args(self):
        presentation_type = self._get_presentation_type()
        
        if presentation_type == '':
            return super().get_args()

        return self._mixin_get_args(presentation_type)

    def get_env(self):
        env = super().get_env()

        presentation_type = self._get_presentation_type()

        if presentation_type != '':
            env = self._mixin_get_env(env, presentation_type)

        return env

    def _expand_user_vars(self, string):
        """
        Expand user related variables in a given string

        Currently expands:
          {USERNAME} -> Name of the user
          {USERID} -> UserID
          {DASHSERVERNAME} -> A Dash plus Server Name (or '' if server is My Server which has no name)
          Also needs to preserve --, port, and base_url to pass on to jhsingle-native-proxy, so don't use straight format function
        """
        dsn = ''
        if self.name:
            dsn = '-{}'.format(self.name)
        return string.replace('{USERNAME}', self.user.name).replace('{USERID}', str(self.user.id)).replace('{DASHSERVERNAME}', dsn)




from tornado.log import app_log

from .builders import Builder, BuildException


class ProcessBuilder(Builder):
 
    async def start(self, dashboard, dashboard_user, db):
        """Start the dashboard

        Returns:
          (str, str): the (new_server_name, new_server_options) of the new dashboard server.

        """

        app_log.info('Starting LocalProcess Builder start function')

        self.event_queue = []

        self.add_progress_event({'progress': 10, 'message': 'Starting builder'})

        self._build_pending = True

        new_server_name = self.format_string(self.cdsconfig.server_name_template)

        if not self.allow_named_servers:
            raise BuildException(400, "Named servers are not enabled.")
        if (
            self.named_server_limit_per_user > 0
            and new_server_name not in dashboard_user.orm_spawners
        ):
            named_spawners = list(dashboard_user.all_spawners(include_default=False))
            if self.named_server_limit_per_user <= len(named_spawners):
                raise BuildException(
                    "User {} already has the maximum of {} named servers."
                    "  One must be deleted before a new server can be created".format(
                        dashboard_user.name, self.named_server_limit_per_user
                    ),
                )
        spawner = dashboard_user.spawners[new_server_name] # Could be orm_spawner or Spawner wrapper

        if spawner.ready:
            # include notify, so that a server that died is noticed immediately
            # set _spawn_pending flag to prevent races while we wait
            spawner._spawn_pending = True
            try:
                state = await spawner.poll_and_notify()
            finally:
                spawner._spawn_pending = False

        new_server_options = {
            'presentation_type': dashboard.presentation_type or 'voila',
            'presentation_path': dashboard.start_path,
            'cmd': ['python3', '-m', 'jhsingle_native_proxy.main'],
            'environment': {
                'JUPYTERHUB_ANYONE': '{}'.format(dashboard.allow_all and '1' or '0'),
                'JUPYTERHUB_GROUP': '{}'.format(dashboard.groupname)
                }
            }

        return (new_server_name, new_server_options)






import re
from datetime import datetime
from collections import defaultdict
from asyncio import sleep, CancelledError

from async_generator import aclosing
from tornado.web import HTTPError
from tornado.log import app_log

from jupyterhub.orm import User
from jupyterhub.utils import iterate_until

from ..util import maybe_future
from ..orm import Dashboard
from ..app import BuildersStore
from ..dbutil import is_upgrade_needed

def check_database_upgrade(f):
    def handler(self, *args, **kwargs):
        engine = self.db.get_bind()
        if is_upgrade_needed(engine):
            return self.redirect("{}hub/dashboards-db-upgrade".format(self.settings['base_url']))
        return f(self, *args, **kwargs)
    return handler


class DashboardBaseMixin:

    unsafe_regex = re.compile(r'[^a-zA-Z0-9]+')

    trailingdash_regex = re.compile(r'\-+$')

    name_regex = re.compile(r'^[a-zA-Z0-9][a-zA-Z0-9_\- \!\@\$\(\)\*\+\?\<\>\.]{0,99}$')

    start_path_regex = re.compile(r'^[A-Za-z0-9\-\._~\:\/\?#\[\]@\!\$\&\(\)\*\+ ,;\%\=]*$')

    def calc_urlname(self, dashboard_name):
        base_urlname = re.sub(self.unsafe_regex, '-', dashboard_name).lower()[:35]

        base_urlname = re.sub(self.trailingdash_regex, '', base_urlname)

        urlname = base_urlname

        self.log.debug('calc safe name from '+urlname)

        now_unique = False
        counter = 1
        while not now_unique:
            orm_dashboard = Dashboard.find(db=self.db, urlname=urlname)
            self.log.info("{} - {}".format(urlname, orm_dashboard))
            if orm_dashboard is None or counter >= 100:
                now_unique = True
            else:
                urlname = "{}-{}".format(base_urlname, counter)
                counter += 1

        self.log.debug('calc safe name : '+urlname)
        return urlname

    @staticmethod
    async def pipe_spawner_progress(dashboard_user, new_server_name, builder):
        
        while True:

            await sleep(0.01)

            if builder._build_future.done():
                break

            if new_server_name in dashboard_user.spawners and dashboard_user.spawners[new_server_name].pending \
                and dashboard_user.spawners[new_server_name]._spawn_future:

                spawner = dashboard_user.spawners[new_server_name]

                app_log.debug('Found spawner for progress')

                async with aclosing(
                    iterate_until(builder._build_future, spawner._generate_progress())
                ) as spawnevents:
                    try:
                        async for event in spawnevents:
                            if 'message' in event:
                                builder.add_progress_event({
                                    'progress': 95, 'message': 'Spawner progress: {}'.format(event['message'])
                                    })
                    except CancelledError:
                        pass

                break 

    def get_source_spawners(self, user):
        return [spawner for spawner in user.all_spawners(include_default=True) if not spawner.orm_spawner.dashboard_final_of]

    def get_visitor_dashboards(self, user):
        orm_dashboards = set()

        for group in user.groups:
            if group.dashboard_visitors_for:
                orm_dashboards.add(group.dashboard_visitors_for)

        orm_dashboards.update(self.db.query(Dashboard).filter(Dashboard.allow_all==True).filter(Dashboard.user != user.orm_user).all())

        user_dashboard_groups = defaultdict(list)

        for dash in orm_dashboards:
            if dash.user: # Just in case it's null due to db corruption
                user_dashboard_groups[dash.user.name].append(dash)

        return user_dashboard_groups

    def get_visitor_users(self, exclude_user_id):
        return self.db.query(User).filter(User.id != exclude_user_id).all()

    def get_visitor_tuples(self, exclude_user_id, existing_group_users=None):
        possible_visitor_users = self.get_visitor_users(exclude_user_id)

        visitor_users = []

        if existing_group_users is not None:
            visitor_users = existing_group_users

        return [(True, user) for user in visitor_users] + [(False, user) for user in set(possible_visitor_users) - set(visitor_users)]

    async def maybe_start_build(self, dashboard, dashboard_user, force_start=False):

        builders_store = BuildersStore.get_instance(self.settings['config'])

        builder = builders_store[dashboard]

        need_follow_progress = True

        def do_final_build(f):
            if f.cancelled() or f.exception() is None:
                builder._build_future = None
            builder._build_pending = False

        if not builder.pending and (dashboard.final_spawner is None or force_start):
            
            if builder._build_future and builder._build_future.done() and builder._build_future.exception() and not force_start:
                pass # Progress stream should display the error for us

            else:
                self.log.debug('starting builder')

                builder._build_pending = True

                async def do_build():

                    # This section would be better if it was inside dockerbuilder, but we want 
                    # to make use of self.spawn_single_user without repeating the code
                    if dashboard.source_spawner and dashboard.source_spawner.name:

                        # Get source spawner
                        source_spawner = dashboard_user.spawners[dashboard.source_spawner.name]

                        if source_spawner.ready:
                            self.log.debug('Source spawner is already ready')
                        elif source_spawner.pending in ['spawn', 'check']:
                            builder.add_progress_event({'progress': 15, 'message': 'Attacheding to pending source server for Dashboard'})
                            self.log.debug('Source spawner is pending - await')
                            spawn_future = final_spawner.getattr('_{}_future'.format(final_spawner.pending), None)
                            if spawn_future:
                                await spawn_future
                        else:
                            builder.add_progress_event({'progress': 10, 'message': 'Starting up source server for Dashboard'})
                            self.log.debug('Source spawner needs a full start')
                            await self.spawn_single_user(dashboard_user, dashboard.source_spawner.name)

                    # Delete existing final spawner if it exists
                    await self.maybe_delete_existing_server(dashboard.final_spawner, dashboard_user)

                    (new_server_name, new_server_options) =  await builder.start(dashboard, dashboard_user, self.db)

                    builder.add_progress_event({'progress': 80, 'message': 'Starting up final server for Dashboard, after build'})

                    maybe_future(self.pipe_spawner_progress(dashboard_user, new_server_name, builder))

                    await self.spawn_single_user(dashboard_user, new_server_name, options=new_server_options)

                    self.db.expire(dashboard) # May have changed during async code above

                    if new_server_name in dashboard_user.orm_user.orm_spawners:
                        dashboard.final_spawner = dashboard_user.orm_user.orm_spawners[new_server_name]

                    # TODO if not, then what?

                    dashboard.started = datetime.utcnow()

                    self.db.commit()

                builder._build_future = maybe_future(do_build())

                builder._build_future.add_done_callback(do_final_build)

        elif builder.pending:
            pass # Progress will show where we've got to

        elif dashboard.final_spawner:
            user = self._user_from_orm(dashboard_user)

            final_spawner = user.spawners[dashboard.final_spawner.name]

            if final_spawner.ready:
                need_follow_progress = False

            elif final_spawner.pending:

                if final_spawner.pending in ['spawn', 'check']:
                    builder._build_pending = True

                    builder.add_progress_event({'progress': 90, 'message': 'Attaching to spawn of final server for Dashboard'})

                    # TODO This branch is rare, but should attach pipe to spawner progress

                    final_spawner.getattr('_{}_future'.format(final_spawner.pending)).add_done_callback(do_final_build)

                else:
                    builder.add_progress_event({'failed': True, 'message': 'Final server for Dashboard is Stopping'})

            else:

                builder._build_pending = True

                builder.event_queue = []

                builder.add_progress_event({'progress': 80, 'message': 'Starting up final server for Dashboard'})

                maybe_future(self.pipe_spawner_progress(user, final_spawner.name, builder))

                builder._build_future = maybe_future(self.spawn_single_user(user, final_spawner.name))

                builder._build_future.add_done_callback(do_final_build)

        return need_follow_progress

    async def maybe_delete_existing_server(self, orm_spawner, dashboard_user):
        if not orm_spawner:
            return
        
        server_name = orm_spawner.name

        user = self._user_from_orm(dashboard_user)

        spawner = user.spawners[server_name]

        if server_name:
            if not self.allow_named_servers:
                raise HTTPError(400, "Named servers are not enabled.")
            if server_name not in user.orm_spawners:
                self.log.debug("%s does not exist anyway", spawner._log_name)
                return
        else:
            raise HTTPError(400, "Cannot delete the default server")

        if spawner.pending == 'stop':
            self.log.debug("%s already stopping", spawner._log_name)

            await spawner._stop_future

        if spawner.ready or spawner.pending == 'spawn':

            if spawner.pending == 'spawn':
                self.log.debug("%s awaiting spawn to finish so can shut it down", spawner._log_name)
                await spawner._spawn_future

            # include notify, so that a server that died is noticed immediately
            status = await spawner.poll_and_notify()
            if status is None:
                stop_future = await self.stop_single_user(user, server_name)
                await stop_future

        self.log.info("Deleting spawner %s", spawner._log_name)
        self.db.delete(spawner.orm_spawner)
        user.spawners.pop(server_name, None)
        self.db.commit()




from jupyterhub.spawner import LocalProcessSpawner
from traitlets import Unicode, Integer

from .variablemixin import VariableMixin


class VariableLocalProcessSpawner(LocalProcessSpawner, VariableMixin):
    
    voila_template = Unicode(
        'materialstream',
        help="""
        --template argument to pass to Voila. Default is materialstream
        """,
    ).tag(config=True)

    proxy_request_timeout = Integer(
        0,
        help="""
        Request timeout in seconds that jhsingle-native-proxy should allow when proxying to the underlying process (e.g. Voila).
        The default of 0 means that no --request-timeout flag will be passed to jhsingle-native-proxy so it will use its own default.
        """,
    ).tag(config=True)

    async def start(self):
        self._pre_start()
        return await super().start()

    def get_args(self):
        presentation_type = self._get_presentation_type()
        
        if presentation_type == '':
            return super().get_args()

        return self._mixin_get_args(presentation_type)

    def get_env(self):
        env = super().get_env()

        presentation_type = self._get_presentation_type()

        if presentation_type != '':
            env = self._mixin_get_env(env, presentation_type)

        return env
